import { Guid } from 'guid-typescript';
import { sumBy, filter, each, sortBy, min, orderBy, forIn, has, remove } from 'lodash';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const FlowMode = {
    None: "",
    ProRata: "ProRata",
    FirstInFirstOut: "FirstInFirstOut",
    LastInFirstOut: "LastInFirstOut",
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const BatchType = {
    Source: "Source",
    Target: "Target",
    Adjustment: "Adjustment",
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const AdjustmentMode = {
    Absolute: "Absolute",
    Relative: "Relative",
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function Criteria() { }
;
/**
 * @record
 */
function IBatch() { }
if (false) {
    /** @type {?|undefined} */
    IBatch.prototype.id;
    /** @type {?|undefined} */
    IBatch.prototype.name;
    /** @type {?} */
    IBatch.prototype.time;
    /** @type {?} */
    IBatch.prototype.type;
    /** @type {?|undefined} */
    IBatch.prototype.criteria;
    /** @type {?|undefined} */
    IBatch.prototype.quantity;
    /** @type {?|undefined} */
    IBatch.prototype.attributes;
    /** @type {?|undefined} */
    IBatch.prototype.measures;
    /** @type {?|undefined} */
    IBatch.prototype.flowMode;
}
class Batch {
    /**
     * @param {?=} batch
     */
    constructor(batch = null) {
        this.id = Guid.create().toString();
        this.criteria = {};
        this.attributes = {};
        this.flowMode = FlowMode.ProRata;
        this.adjustmentMode = AdjustmentMode.Absolute;
        this.flowsIn = [];
        this.flowsOut = [];
        /**
         * returns the balance of the batch
         */
        this.balanceAtTime = (/**
         * @param {?=} time
         * @return {?}
         */
        (time) => {
            //  Check whether the time is before the time of this batch
            if (time < this.time)
                return null;
            /** @type {?} */
            let balance = 0;
            //  Source batches have an intrinsic quantity that does not come from FlowsIn
            if (this.type == BatchType.Source)
                balance += this.quantity;
            balance = balance
                + sumBy(this.flowsIn, (/**
                 * @param {?} x
                 * @return {?}
                 */
                x => x.quantity))
                - sumBy(filter(this.flowsOut, (/**
                 * @param {?} x
                 * @return {?}
                 */
                x => x.target.time <= time)), (/**
                 * @param {?} x
                 * @return {?}
                 */
                x => x.quantity));
            return balance;
        });
        this.balancesOverTime = (/**
         * @param {?} times
         * @return {?}
         */
        (times) => {
            /** @type {?} */
            let balances = [];
            each(times, (/**
             * @param {?} t
             * @return {?}
             */
            t => {
                balances.push(this.balanceAtTime(t));
            }));
            return balances;
        });
        //  Returns the calculated attributes of this batch at the given time
        this.attributesAtTime = (/**
         * @param {?=} time
         * @return {?}
         */
        (time) => {
            //  Check whether the time is valid
            if (time < this.time)
                return null;
            //  Start with the target attributes of this batch
            return this.attributes;
        });
        this.attributesOverTime = (/**
         * @param {?} times
         * @return {?}
         */
        (times) => {
            /** @type {?} */
            let attributes = [];
            each(times, (/**
             * @param {?} t
             * @return {?}
             */
            t => {
                attributes.push(this.attributesAtTime(t));
            }));
            return attributes;
        });
        Object.assign(this, batch);
    }
    /**
     * @return {?}
     */
    get time() {
        return this._time;
    }
    /**
     * @param {?} value
     * @return {?}
     */
    set time(value) {
        if (value instanceof Date) {
            this._time = value;
        }
        else {
            this._time = new Date(value);
        }
    }
}
if (false) {
    /** @type {?} */
    Batch.prototype.type;
    /** @type {?} */
    Batch.prototype.id;
    /** @type {?} */
    Batch.prototype.name;
    /**
     * @type {?}
     * @private
     */
    Batch.prototype._time;
    /** @type {?} */
    Batch.prototype.criteria;
    /** @type {?} */
    Batch.prototype.quantity;
    /** @type {?} */
    Batch.prototype.attributes;
    /** @type {?} */
    Batch.prototype.measures;
    /** @type {?} */
    Batch.prototype.flowMode;
    /** @type {?} */
    Batch.prototype.adjustmentMode;
    /** @type {?} */
    Batch.prototype.flowsIn;
    /** @type {?} */
    Batch.prototype.flowsOut;
    /**
     * returns the balance of the batch
     * @type {?}
     */
    Batch.prototype.balanceAtTime;
    /** @type {?} */
    Batch.prototype.balancesOverTime;
    /** @type {?} */
    Batch.prototype.attributesAtTime;
    /** @type {?} */
    Batch.prototype.attributesOverTime;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class Flow {
}
if (false) {
    /** @type {?} */
    Flow.prototype.source;
    /** @type {?} */
    Flow.prototype.target;
    /** @type {?} */
    Flow.prototype.time;
    /** @type {?} */
    Flow.prototype.quantity;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class Lineage {
    constructor() {
        this.batches = [];
        this.flows = [];
        this.add = (/**
         * @param {?} batch
         * @return {?}
         */
        (batch) => {
            //  Convert to a Batch object if required
            if (batch instanceof Batch) {
                this.batches.push(batch);
                return batch;
            }
            else {
                return this.add(new Batch(batch));
            }
        });
        this.createFlow = (/**
         * @param {?} source
         * @param {?} target
         * @param {?} time
         * @param {?} quantity
         * @return {?}
         */
        (source, target, time, quantity) => {
            /** @type {?} */
            let flow = new Flow();
            flow.source = source;
            flow.target = target;
            flow.time = time;
            flow.quantity = quantity;
            source.flowsOut.push(flow);
            target.flowsIn.push(flow);
            this.flows.push(flow);
        });
    }
    /**
     * @return {?}
     */
    process() {
        //  Clear out current flows
        each(this.batches, (/**
         * @param {?} x
         * @return {?}
         */
        x => {
            x.flowsIn.length = 0;
            x.flowsOut.length = 0;
        }));
        this.flows.length = 0;
        //  Filter out the batches that do not need processing (ie. the Sources)
        /** @type {?} */
        let batchesToProcess = filter(this.batches, (/**
         * @param {?} x
         * @return {?}
         */
        x => x.type === BatchType.Target || x.type === BatchType.Adjustment));
        //  Order by Time
        batchesToProcess = sortBy(batchesToProcess, (/**
         * @param {?} batch
         * @return {?}
         */
        batch => batch.time));
        //  Process each batch
        each(batchesToProcess, (/**
         * @param {?} target
         * @return {?}
         */
        target => {
            //  Determine the type of target and process accordingly
            if (target.type === BatchType.Adjustment) {
                this.processAdjustment(target);
            }
            else if (target.type === BatchType.Target) {
                //  Find candidate source batches
                /** @type {?} */
                let candidates = this.findCandidates(target);
                //  Process the batch
                switch (target.flowMode) {
                    case FlowMode.ProRata:
                        this.processProRata(target, candidates);
                        break;
                    case FlowMode.FirstInFirstOut:
                        this.processFirstInFirstOut(target, candidates);
                        break;
                    case FlowMode.LastInFirstOut:
                        this.processLastInFirstOut(target, candidates);
                        break;
                }
            }
        }));
    }
    /**
     * @param {?} adjustment
     * @return {?}
     */
    processAdjustment(adjustment) {
        //  Process Adjustment according to the AdjustmentMode
        if (adjustment.adjustmentMode == AdjustmentMode.Absolute) {
            this.processAbsoluteAdjustment(adjustment);
        }
    }
    /**
     * @param {?} adjustment
     * @return {?}
     */
    processAbsoluteAdjustment(adjustment) {
        //  Find the source batches to adjust
        /** @type {?} */
        let candidates = this.findCandidates(adjustment);
        //  Find the total quantity of all the candidates
        /** @type {?} */
        let totalQuantity = sumBy(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => x.balanceAtTime(adjustment.time)));
        //  Determine the relative adjustment, and negate it as we are flowing OUT of the sources rather than in
        /** @type {?} */
        let relativeAdjustment = -1 * (adjustment.quantity - totalQuantity);
        //  Create the flows
        this.createFlowsProRata(candidates, adjustment, relativeAdjustment);
    }
    /**
     * @param {?} sources
     * @param {?} target
     * @param {?} totalFlowQuantity
     * @return {?}
     */
    createFlowsProRata(sources, target, totalFlowQuantity) {
        //  Calculate the existing total quantity available from matching candidates
        /** @type {?} */
        let totalSourcesQuantity = sumBy(sources, (/**
         * @param {?} c
         * @return {?}
         */
        c => c.balanceAtTime(target.time)));
        //  Add flow for each candidate
        each(sources, (/**
         * @param {?} source
         * @return {?}
         */
        source => {
            /** @type {?} */
            let sourceRemainingQuantity = source.balanceAtTime(target.time);
            /** @type {?} */
            let flowQuantity = min([sourceRemainingQuantity, (sourceRemainingQuantity / totalSourcesQuantity) * totalFlowQuantity]);
            this.createFlow(source, target, target.time, flowQuantity);
        }));
    }
    /**
     * @param {?} target
     * @param {?} candidates
     * @return {?}
     */
    processProRata(target, candidates) {
        //  Calculate the total quantity available from matching candidates
        /** @type {?} */
        let totalAvailable = sumBy(candidates, (/**
         * @param {?} c
         * @return {?}
         */
        c => c.balanceAtTime(target.time)));
        //  Add flow for each candidate
        each(candidates, (/**
         * @param {?} source
         * @return {?}
         */
        source => {
            /** @type {?} */
            let sourceRemainingQuantity = source.balanceAtTime(target.time);
            /** @type {?} */
            let flowQuantity = min([sourceRemainingQuantity, (sourceRemainingQuantity / totalAvailable) * target.quantity]);
            this.createFlow(source, target, target.time, flowQuantity);
        }));
    }
    /**
     * @param {?} target
     * @param {?} candidates
     * @return {?}
     */
    processLastInFirstOut(target, candidates) {
        //  Sort the candidates by time, descending
        candidates = orderBy(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => x.time), 'desc');
        /** @type {?} */
        let targetQuantity = target.quantity;
        //  Loop through the sources and create flows until the target is fulfilled or the sources are exhausted
        each(candidates, (/**
         * @param {?} source
         * @return {?}
         */
        source => {
            /** @type {?} */
            let sourceQuantity = source.balanceAtTime(target.time);
            /** @type {?} */
            let flowQuantity = min([targetQuantity, sourceQuantity]);
            this.createFlow(source, target, target.time, flowQuantity);
            targetQuantity -= flowQuantity;
            if (targetQuantity == 0) {
                return false;
            }
        }));
    }
    /**
     * @param {?} target
     * @param {?} candidates
     * @return {?}
     */
    processFirstInFirstOut(target, candidates) {
        //  Sort the candidates by time, descending
        candidates = orderBy(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => x.time), 'asc');
        /** @type {?} */
        let targetQuantity = target.quantity;
        //  Loop through the sources and create flows until the target is fulfilled or the sources are exhausted
        each(candidates, (/**
         * @param {?} source
         * @return {?}
         */
        source => {
            /** @type {?} */
            let sourceQuantity = source.balanceAtTime(target.time);
            /** @type {?} */
            let flowQuantity = min([targetQuantity, sourceQuantity]);
            this.createFlow(source, target, target.time, flowQuantity);
            targetQuantity -= flowQuantity;
            if (targetQuantity == 0) {
                return false;
            }
        }));
    }
    /**
     * returns a set of batches which could potentially fulfil a target batch
     * @param {?} target
     * @return {?}
     */
    findCandidates(target) {
        /** @type {?} */
        let candidates = this.batches;
        //  Eliminate the target from the candidates (a batch cannot fulfil itself)
        candidates = filter(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => x != target));
        //  Eliminate any batches that postdate the target (batches cannot fulfil earlier batches)
        candidates = filter(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => x.time <= target.time));
        //  Eliminates any candidates with zero balance
        //  TODO: optimise this using 'currentBalance' running variable
        candidates = filter(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => x.balanceAtTime(target.time) > 0));
        //  Filter out sources that do not match the target's sources property
        candidates = filter(candidates, (/**
         * @param {?} x
         * @return {?}
         */
        x => this.matchProperties(x, target)));
        return candidates;
    }
    /**
     * @param {?} source
     * @param {?} target
     * @return {?}
     */
    matchProperties(source, target) {
        //  Test for properties
        /** @type {?} */
        let propertyMatched = true;
        forIn(target.criteria, (/**
         * @param {?} value
         * @param {?} key
         * @return {?}
         */
        (value, key) => {
            if (value != null) {
                if (!(has(source.attributes, key) && source.attributes[key] == value)) {
                    propertyMatched = false;
                    return false;
                }
            }
            else {
                if (has(source.attributes, key) && source.attributes[key] != value) {
                    propertyMatched = false;
                    return false;
                }
            }
        }));
        return propertyMatched;
    }
    /**
     * @param {?} batch
     * @return {?}
     */
    remove(batch) {
        remove(this.batches, batch);
    }
}
if (false) {
    /** @type {?} */
    Lineage.prototype.batches;
    /** @type {?} */
    Lineage.prototype.flows;
    /** @type {?} */
    Lineage.prototype.add;
    /** @type {?} */
    Lineage.prototype.createFlow;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { Batch, BatchType, Flow, FlowMode, Lineage };
//# sourceMappingURL=seers-lineage.js.map
